class Mysql
    # Information about one column.
    class Field
        attr_reader :name
        attr_reader :org_name
        attr_reader :table
        attr_reader :org_table
        attr_reader :db
        attr_reader :catalog
        attr_reader :def
        attr_reader :length
        attr_reader :max_length
        attr_reader :name_length
        attr_reader :org_name_length
        attr_reader :table_length
        attr_reader :org_table_length
        attr_reader :db_length
        attr_reader :catalog_length
        attr_reader :def_length
        attr_reader :flags
        attr_reader :decimals
        attr_reader :charsetnr
        attr_reader :type

        def initialize(c_field)
            @name =             c_field[:name]
            @org_name =         c_field[:org_name]
            @table =            c_field[:table]
            @org_table =        c_field[:org_table]
            @db =               c_field[:db]
            @catalog =          c_field[:catalog]
            @def =              c_field[:def]
            @length =           c_field[:length]
            @max_length =       c_field[:max_length]
            @name_length =      c_field[:name_length]
            @org_name_length =  c_field[:org_name_length]
            @table_length =     c_field[:table_length]
            @org_table_length = c_field[:org_table_length]
            @db_length =        c_field[:db_length]
            @catalog_length =   c_field[:catalog_length]
            @def_length =       c_field[:def_length]
            @flags =            c_field[:flags]
            @decimals =         c_field[:decimals]
            @charsetnr =        c_field[:charsetnr]
            # @type =             C::FieldType[c_field[:type]]
            @type =             c_field[:type]
            @flags |= NUM_FLAG if is_num_type?
        end

        # @return [Boolean] true if this field is not null
        def is_not_null?
            @flags & NOT_NULL_FLAG != 0
        end

        # @return [Boolean] true if this is a numeric field
        def is_num?
            @flags & NUM_FLAG != 0
        end

        # @return [Boolean] true if this field is a primary key
        def is_pri_key?
            @flags & PRI_KEY_FLAG != 0
        end

        # @return [Hash] this field as a hash
        def hash
            h = {}
            h['name'] = @name
            h['table'] = @table
            h['def'] = @def
            h['length'] = @length
            h['max_length'] = @max_length
            h['flags'] = @flags
            h['decimals'] = @decimals
            h['type'] = @type
            h
        end

        def inspect
            "#<Mysql::Field:#{@name}>"
        end

        private
        # @return [Boolean] true if this field's type is a numeric type
        def is_num_type?
            [TYPE_DECIMAL, TYPE_TINY, TYPE_SHORT, TYPE_LONG, TYPE_FLOAT, TYPE_DOUBLE, TYPE_LONGLONG, TYPE_INT24].include?(@type) || (@type == TYPE_TIMESTAMP && (@length == 14 || @length == 8))
        end
    end
end
