require 'ffi'

# Basic MySQL class, provides interface to a server.
# @author Frank Fischer
class Mysql

    # FFI interface.
    module C
		extend FFI::Library
		ffi_lib ["mysqlclient",
				 "libmysqlclient.so.15",
				 "libmysqlclient.so.16",
				 "libmysqlclient.so.17",
				 "libmysqlclient.so.18",
				]

#        FieldType = enum(:decimal,     Mysql::Field::TYPE_DECIMAL,
#                         :tiny,        Mysql::Field::TYPE_TINY,
#                         :short,       Mysql::Field::TYPE_SHORT,
#                         :long,        Mysql::Field::TYPE_LONG,
#                         :float,       Mysql::Field::TYPE_FLOAT,
#                         :double,      Mysql::Field::TYPE_DOUBLE,
#                         :null,        Mysql::Field::TYPE_NULL,
#                         :timestamp,   Mysql::Field::TYPE_TIMESTAMP,
#                         :longlong,    Mysql::Field::TYPE_LONGLONG,
#                         :int24,       Mysql::Field::TYPE_INT24,
#                         :date,        Mysql::Field::TYPE_DATE,
#                         :time,        Mysql::Field::TYPE_TIME,
#                         :datetime,    Mysql::Field::TYPE_DATETIME,
#                         :year,        Mysql::Field::TYPE_YEAR,
#                         :newdate,     Mysql::Field::TYPE_NEWDATE,
#                         :varchar,     Mysql::Field::TYPE_VARCHAR,
#                         :bit,         Mysql::Field::TYPE_BIT,
#                         :newdecimal,  Mysql::Field::TYPE_NEWDECIMAL,
#                         :enum,        Mysql::Field::TYPE_ENUM,
#                         :set,         Mysql::Field::TYPE_SET,
#                         :tiny_blob,   Mysql::Field::TYPE_TINY_BLOB,
#                         :medium_blob, Mysql::Field::TYPE_MEDIUM_BLOB,
#                         :long_blob,   Mysql::Field::TYPE_LONG_BLOB,
#                         :blob,        Mysql::Field::TYPE_BLOB,
#                         :var_string,  Mysql::Field::TYPE_VAR_STRING,
#                         :string,      Mysql::Field::TYPE_STRING,
#                         :geometry,    Mysql::Field::TYPE_GEOMETRY,
#                         :char,        Mysql::Field::TYPE_CHAR,
#                         :interval,    Mysql::Field::TYPE_INTERVAL)
        FieldType = :uchar

        class Field < FFI::Struct
            layout(:name,             :string,
                   :org_name,         :string,
                   :table,            :string,
                   :org_table,        :string,
                   :db,               :string,
                   :catalog,          :string,
                   :def,              :string,
                   :length,           :ulong,
                   :max_length,       :ulong,
                   :name_length,      :uint,
                   :org_name_length,  :uint,
                   :table_length,     :uint,
                   :org_table_length, :uint,
                   :db_length,        :uint,
                   :catalog_length,   :uint,
                   :def_length,       :uint,
                   :flags,            :uint,
                   :decimals,         :uint,
                   :charsetnr,        :uint,
                   :type,             FieldType)
        end

        def self.maybe_attach_function name, *args
            attach_function name, *args
        rescue FFI::NotFoundError => e
	        c = class << self; self; end
	        c.class_eval do
		        define_method(name) {|*a| raise e}
	        end
        rescue => e
	        c = class << self; self; end
	        c.class_eval do
		        define_method(name) {|*a| raise e}
	        end
        end

        StmtAttrType = enum( :update_max_length, :cursor_type, :prefetch_rows )
        attach_function :mysql_init, [:pointer], :pointer
        attach_function :mysql_close, [:pointer], :void
        attach_function :mysql_errno, [:pointer], :uint
        attach_function :mysql_error, [:pointer], :string
        attach_function :mysql_warning_count, [:pointer], :uint
        attach_function :mysql_debug, [:string], :void
        attach_function :mysql_dump_debug_info, [:pointer], :int
        attach_function :mysql_ping, [:pointer], :int
        attach_function :mysql_refresh, [:pointer, :uint ], :int
        maybe_attach_function :mysql_reload, [:pointer], :int
        attach_function :mysql_shutdown, [:pointer], :int
        attach_function :mysql_change_user, [:pointer, :string, :string, :string], :char
        attach_function :mysql_select_db, [:pointer, :string], :int
        attach_function :mysql_thread_id, [:pointer], :ulong
        attach_function :mysql_kill, [:pointer, :ulong], :int
        attach_function :mysql_info, [:pointer], :string
        attach_function :mysql_stat, [:pointer], :string
        attach_function :mysql_character_set_name, [:pointer], :string
        attach_function :mysql_set_character_set, [:pointer, :string], :int
        attach_function :mysql_get_client_version, [], :int
        attach_function :mysql_get_client_info, [], :string
        attach_function :mysql_get_server_version, [:pointer], :int
        attach_function :mysql_get_server_info, [:pointer], :string
        attach_function :mysql_get_host_info, [:pointer], :string
        attach_function :mysql_get_proto_info, [:pointer], :uint
        attach_function :mysql_ssl_set, [:pointer, :string, :string, :string, :string, :string], :int
        attach_function :mysql_real_connect, [:pointer, :string, :string, :string, :string, :uint, :string, :ulong], :pointer
        attach_function :mysql_ping, [:pointer], :int
        attach_function :mysql_options, [:pointer, :int, :pointer], :int
        attach_function :mysql_set_server_option, [:pointer, :int], :int
        attach_function :mysql_real_escape_string, [:pointer, :pointer, :char, :ulong], :ulong
        attach_function :mysql_real_query, [:pointer, :string, :ulong], :int
        attach_function :mysql_insert_id, [:pointer], :ulong_long
        attach_function :mysql_list_dbs, [:pointer, :string], :pointer
        attach_function :mysql_list_fields, [:pointer, :string, :string], :pointer
        attach_function :mysql_list_processes, [:pointer], :pointer
        attach_function :mysql_list_tables, [:pointer, :string], :pointer
        attach_function :mysql_autocommit, [:pointer, :char], :char
        attach_function :mysql_commit, [:pointer], :char
        attach_function :mysql_rollback, [:pointer], :char
        attach_function :mysql_field_count, [:pointer], :uint
        attach_function :mysql_store_result, [:pointer], :pointer
        attach_function :mysql_use_result, [:pointer], :pointer
        attach_function :mysql_free_result, [:pointer], :void
        attach_function :mysql_next_result, [:pointer], :int
        attach_function :mysql_more_results, [:pointer], :bool
        attach_function :mysql_affected_rows, [:pointer], :ulong_long
        attach_function :mysql_num_rows, [:pointer], :ulong_long
        attach_function :mysql_fetch_row, [:pointer], :pointer
        attach_function :mysql_fetch_lengths, [:pointer], :pointer
        attach_function :mysql_row_tell, [:pointer], :ulong_long
        attach_function :mysql_row_seek, [:pointer, :ulong_long], :ulong_long
        attach_function :mysql_num_fields, [:pointer], :uint
        attach_function :mysql_fetch_field, [:pointer], :pointer
        attach_function :mysql_fetch_field_direct, [:pointer, :uint], :pointer
        attach_function :mysql_field_tell, [:pointer], :uint
        attach_function :mysql_field_seek, [:pointer, :uint], :uint
        attach_function :mysql_data_seek, [:pointer, :ulong_long], :void
        attach_function :mysql_sqlstate, [:pointer], :string
        attach_function :mysql_stmt_init, [:pointer], :pointer
        attach_function :mysql_stmt_attr_set, [:pointer, StmtAttrType, :pointer], :int
        attach_function :mysql_stmt_close, [:pointer], :void
        attach_function :mysql_stmt_prepare, [:pointer, :string, :ulong], :int
        attach_function :mysql_stmt_execute, [:pointer], :int
        attach_function :mysql_stmt_free_result, [:pointer], :int
        attach_function :mysql_stmt_param_count, [:pointer], :ulong
        attach_function :mysql_stmt_bind_param, [:pointer, :pointer], :char
        attach_function :mysql_stmt_result_metadata, [:pointer], :pointer
        attach_function :mysql_stmt_bind_result, [:pointer, :pointer], :char
        attach_function :mysql_stmt_affected_rows, [:pointer], :ulong_long
        attach_function :mysql_stmt_store_result, [:pointer], :int
        attach_function :mysql_stmt_fetch, [:pointer], :int
        attach_function :mysql_stmt_data_seek, [:pointer, :ulong_long], :void
        attach_function :mysql_stmt_field_count, [:pointer], :uint
        attach_function :mysql_stmt_num_rows, [:pointer], :ulong_long
        attach_function :mysql_stmt_row_tell, [:pointer], :ulong_long
        attach_function :mysql_stmt_row_seek, [:pointer, :ulong_long], :ulong_long
        attach_function :mysql_stmt_insert_id, [:pointer], :ulong_long
        attach_function :mysql_stmt_sqlstate, [:pointer], :string
        attach_function :mysql_stmt_errno, [:pointer], :uint
        attach_function :mysql_stmt_error, [:pointer], :string

    end

    # Creates a new MySQL object.
    def self.init
        mysql = allocate
        mysql.send :initialize
        mysql
    end

    # Creates a new MySQL connector and opens a connection.
    def self.new( host = nil, user = nil, passwd = nil, db = nil, port = 0, sock = nil, flag = 0)
        mysql = allocate
        mysql.send :initialize
        mysql.real_connect( host, user, passwd, db, port, sock, flag )
        mysql
    end

    # @return [Integer] the version of the client
    def self.client_version
        C::mysql_get_client_version
    end

    # @return [Integer] the version of the client
    def client_version
        Mysql.client_version
    end
    alias get_client_version client_version

    # @return [String] string containing the client's version
    def self.client_info
        C::mysql_get_client_info
    end

    # @return [String] string containing the client's version
    def client_info
        Mysql.client_info
    end
    alias get_client_info client_info

    # Does a DBUG_PUSH with the given string.
    def self.debug(arg)
        C::mysql_debug(arg)
        self
    end

    # Escape special character in MySQL.
    # === Note
    # In Ruby 1.8, this is not safe for multibyte charset such as 'SJIS'.
    # You should use place-holder in prepared-statement.
    def self.escape_string(str)
        str.gsub(/[\0\n\r\\\'\"\x1a]/) do |s|
            case s
            when "\0" then "\\0"
            when "\n" then "\\n"
            when "\r" then "\\r"
            when "\x1a" then "\\Z"
            else "\\#{s}"
            end
        end
    end

    class << self
        alias :real_connect :new
        alias :connect :new
        alias :get_client_version :client_version
        alias :get_client_info :client_info
        alias quote escape_string
    end


    # if true (the default), query return the first result
    attr_accessor :query_with_result

    # Creates a new connection to a MySQL server.
    #
    # @param (see Mysql#real_connect)
    def initialize
        @mysql_free = [true]
        @mysql = C::mysql_init( nil )
        @connected = false
        @query_with_result = true
        @reconnect = false
    end

    # internal finalizer
    def self.finalizer(mysql, mysql_free)
        Proc.new do |*args|
            unless mysql_free[0]
                C::mysql_close(mysql)
            end
        end
    end

    # Set SSL settings (must be called before real_connect)
    #
    # @param [String] key path to the key file
    # @param [String] cert path to the certificate file
    # @param [String] ca path to the certificate authority file
    # @param [String] capath path name to a directory with SSL CA certificates in pem format
    # @param [String] cipher list of permissible ciphers to use
    #
    # Any unused SSL parameters may be given as NULL.
    def ssl_set(key, cert, ca, capath, cipher)
      if C::mysql_ssl_set(@mysql, key, cert, ca, capath, cipher) != 0
        raise Error.new(error, errno)
      end
      self
    end

    # Opens a new connection to a MySQL server.
    #
    # @param [String] host the MySQL server
    # @param [String] user the username to login
    # @param [String] passwd the user's password
    # @param [String] db the name of the database to use
    # @param [Integer] port the port of the server to use
    # @param [Integer] flag connection flags
    def real_connect( host = nil, user = nil, passwd = nil, db = nil, port = 0, sock = nil, flag = 0)
        raise Error, "Already connected to a MySQL server" if @connected

        ObjectSpace.define_finalizer( self, Mysql.finalizer(@mysql, @mysql_free))

        ret = C::mysql_real_connect( @mysql, host, user, passwd, db, port, sock, flag )
        if ret.nil? or ret.null?
            raise Mysql::Error, error
        end

        @connected = true
        self
    end
    alias :connect :real_connect

    # Closes the connection to the server.
    def close
        C::mysql_close(@mysql)
        @mysql = nil
        @mysql_free[0] = true
        @connected = false
        self
    end

    # Checks whether the connection to the server is working.
    def ping
        if C::mysql_ping(@mysql) != 0
            raise Error.new(error, errno)
        end
        self
    end

    # @return [Integer] error code for the most recently failed API function
    def errno
        C::mysql_errno(@mysql)
    end

    # @return [String] error message for the most recently failing API function
    def error
        C::mysql_error(@mysql)
    end

    # @return [Integer] the number of errors, warnings and notes generated during the previous SQL statement
    def warning_count
        C::mysql_warning_count(@mysql)
    end

    # Write debug information to the log
    # Raises an error when there is a problem, otherwise returns self.
    def dump_debug_info
        if C::mysql_dump_debug_info(@mysql) != 0
            raise Error.new(error, errno)
        end
        self
    end

    # Check whether the connection to the server is working.
    # Raises an error when there is a problem, otherwise returns self.
    def ping
        if C::mysql_ping(@mysql) != 0
            raise Error.new(error, errno)
        end
        self
    end

    # Flush tables or caches, or reset replication server information.
    #
    # @param [Integer] options bit mask of REFRESH_... values
    def refresh(options)
        if C::mysql_refresh(@mysql, options) != 0
            raise Error.new(error, errno)
        end
        self
    end

    # Asks the MySQL server to reload the grant table.
    def reload
        if C::mysql_reload(@mysql) != 0
            raise Error.new(error, errno)
        end
        self
    end

    # Asks the MySQL server to shut down.
    def shutdown
        if C::mysql_shutdown(@mysql) != 0
            raise Error.new(error, errno)
        end
        self
    end

    # Change the user and change to a different database
    #
    # @param [String] user user to change to
    # @param [String] password password of user
    # @param [String] db name of database to change to
    #
    # Raises an error when there is a problem, otherwise returns self.
    def change_user(user, password, db)
        if C::mysql_change_user(@mysql, user, password, db) != 0
            raise Error.new(error, errno)
        end
        self
    end

    # Change the current database
    #
    # @param [String] db name of database to change to
    #
    # Raises an error when there is a problem, otherwise returns self.
    def select_db(db)
        if C::mysql_select_db(@mysql, db) != 0
            raise Error.new(error, errno)
        end
        self
    end

    # Returns the thread ID of the current connection
    # @return [Integer] thread id of the current connection
    def thread_id
        C::mysql_thread_id(@mysql)
    end

    # Asks the server to kill the specified thread.
    # @param [Integer] pid thread id to kill
    def kill(pid)
        if C::mysql_kill(@mysql, pid) != 0
            raise Error.new(error, errno)
        end
        self
    end

    # Return information about the most recently executed statement
    # @return [String] information about statement
    def info
        C::mysql_info(@mysql)
    end

    # Return information about the server status
    # @return [String] string describing the server status
    def stat
        C::mysql_stat(@mysql)
    end

    # @return [String] default character set name for the current connection
    def character_set_name
        C::mysql_character_set_name(@mysql)
    end

    # Set default character set for the current connection.
    # @param [String] csname default character set to use
    def set_character_set(csname)
        if C::mysql_set_character_set(@mysql, csname) != 0
            raise Error.new(error, errno)
        end
        self
    end

    # Returns a legal SQL string to be used in SQL statements
    # @param [String] str The string to be escaped
    # @return [String] The escaped string
    def real_escape_string(str)
        to = FFI::MemoryPointer.new(str.size * 2 + 1, 1, false)
        len = C::mysql_real_escape_string(@mysql, to, str, str.size)
        to.read_string(len)
    end
    alias :escape_string :real_escape_string
    alias :quote :real_escape_string

    # @return [Integer] the version of the server
    def server_version
        C::mysql_get_server_version(@mysql)
    end
    alias get_server_version server_version

    # @return [String] string containing the server's version
    def server_info
        C::mysql_get_server_info(@mysql)
    end
    alias get_server_info server_info

    # @return [String] string containing the host name and connection type
    def host_info
        C::mysql_get_host_info(@mysql)
    end
    alias get_host_info host_info

    # @return [Integer] containing the current connection's procol version
    def proto_info
        C::mysql_get_proto_info(@mysql)
    end
    alias get_proto_info proto_info

    # @return [String] the SQLSTATE error code for the most recent statement
    def sqlstate
        C::mysql_sqlstate(@mysql)
    end

    # Sets extra connection options.
    #
    # @param [Integer] option the option to set
    # @param [String,Integer,true,false,nil] the value of the option to set
    def options( arg, value = nil )
        result = if value.nil?
                     C::mysql_options( @mysql, arg, nil )
                 elsif value.kind_of? Integer
                     C::mysql_options( @mysql, arg, FFI::MemoryPointer.new(:uint).write_int(value) )
                 elsif value.kind_of? String
                     C::mysql_options( @mysql, arg, FFI::MemoryPointer.from_string(value) )
                 elsif value == true
                     C::mysql_options( @mysql, arg, FFI::MemoryPointer.new(:uint).write_int(1) )
                 elsif value == false
                     C::mysql_options( @mysql, arg, FFI::MemoryPointer.new(:uint).write_int(0) )
                 else
                     raise ArgumentError, "value must one of [String, Integer, nil, true, false]"
                 end
        raise Error.new(error, errno) if result != 0
        self
    end

    # Enables or disables an options for the connection.
    #
    # @param [Integer] option server option
    def set_server_option( option )
        if C::mysql_set_server_option( @mysql, option ) != 0
            raise Error.new(error, errno)
        end
        self
    end

    # Sets the auto-reconnect flag of mysql
    #
    # @param [true,false] The new reconnect state.
    def reconnect= flag
        options OPT_RECONNECT, flag
        @reconnect = flag
    end

    # Returns the current state of the reconnect flag
    # @return [true,false] true if auto-reconnection is enabled.
    def reconnect
        @reconnect
    end

    # Returns the value generated for an AUTO_INCREMENT column by
    # the previous INSERT or UPDATE statement.
    def insert_id
        C::mysql_insert_id(@mysql)
    end

    # Returns a list of all databases on the server matching a pattern
    # @param [String,nil] wild The pattern the databases must match
    def list_dbs(wild = nil)
        res = C::mysql_list_dbs(@mysql, wild)
        if res.nil? or res.null?
            raise Error.new(error, errno)
        end
        res = Result.new(@mysql, res)
        dbs = []
        res.each do |row|
            dbs << row[0]
        end
        dbs
    end

    # Returns a result consisting the field names in the given table matching a pattern
    # @param [String] table The table whose fields should be returned
    # @param [String,nil] wild The pattern the field names must match
    # @return [Result] A MySQL result of the field names
    def list_fields(table, wild = nil)
        res = C::mysql_list_fields(@mysql, table, wild)
        if res.nil? or res.null?
            raise Error.new(error, errno)
        end
        Result.new(@mysql, res)
    end

    # Returns a result describing the current server threads
    # @return [Result] A MySQL result of the current server threads
    def list_processes
        res = C::mysql_list_processes(@mysql)
        if res.nil? or res.null?
            raise Error.new(error, errno)
        end
        Result.new(@mysql, res)
    end

    # Returns a result consisting the table names matching a pattern
    # @param [String,nil] wild The pattern the table names must match
    # @return [Array<String>] The table names
    def list_tables(wild = nil)
        res = C::mysql_list_tables(@mysql, wild)
        if res.nil? or res.null?
            raise Error.new(error, errno)
        end
        res = Result.new(@mysql, res)
        tables = []
        res.each do |row|
            tables << row[0]
        end
        tables
    end

    # Sets autocommit mode on if mode is true, off if mode is false.
    def autocommit(mode)
      if C::mysql_autocommit(@mysql, mode ? 1 : 0) != 0
        raise Error.new(error, errno)
      end
      self
    end

    # Commit transaction
    def commit
        if C::mysql_commit(@mysql) != 0
            raise Error.new(error, errno)
        end
        self
    end

    # Rollback the current transaction
    def rollback
        if C::mysql_rollback(@mysql) != 0
            raise Error.new(error, errno)
        end
        self
    end

    # Execute a query statement.
    #
    # @param [String] sql the SQL statement
    # @yield [optional, Result] calls block once per result set
    #
    # @return [Result,self] the first result set if no block is given, self otherwise
    def query(sql)
        raise Error, "Not connected" unless @connected
        if C::mysql_real_query(@mysql, sql, sql.size) != 0
            raise Error.new(error, errno)
        end

        if block_given?
            begin
                result = store_result
                yield result
            ensure
                result.free
            end while next_result
            self
        elsif query_with_result
            if field_count == 0
                nil
            else
                store_result
            end
        else
            self
        end
    end
    alias real_query query

    # Stores the current result in a result set.
    # @return [Result] the result set
    def store_result
        res = C::mysql_store_result(@mysql)
        if res.nil? or res.null?
            raise Error.new(error, errno)
        end
        Result.new(@mysql, res)
    end

    # Initialize result set retrieval but do not read data.
    # @return [Result] the result set
    def use_result
        res = C::mysql_use_result(@mysql)
        if res.nil? or res.null?
            raise Error.new(error, errno)
        end
        Result.new(@mysql, res)
    end

    # Advances to the next result set.
    # @return [true,false] true if there's another result set
    def next_result
        result = C::mysql_next_result(@mysql)
        if result == 0
            true
        elsif result < 0
            false
        else
            raise Error.new(error, errno)
        end
    end

    # @return [true,false] true if there's another result set
    def more_results?
        C::mysql_more_results(@mysql)
    end
    alias more_results more_results?

    # @return [Integer] the number of affected rows by the last query
    def affected_rows
         C::mysql_affected_rows(@mysql)
    end

    # @return [Integer] the number of columns for the most recent query
    def field_count
        C::mysql_field_count(@mysql)
    end

    # @return [Stmt] a new statement
    def stmt_init
        Stmt.new( @mysql )
    end

    # Creates and prepares a new statement.
    # @param [String] stmt the SQL statement
    # @return [Stmt] the new prepared statement
    def prepare( stmt )
        stmt_init.prepare(stmt)
    end
end
